from google.cloud.bigquery import RangePartitioning, PartitionRange, TimePartitioning

from skt.ye import get_spark
from google.cloud.bigquery.job import QueryJobConfig
from google.cloud.exceptions import NotFound


def _bq_cell_magic(line, query):
    from IPython.core import magic_arguments
    from google.cloud.bigquery.magics import _cell_magic
    import time

    start = time.time()
    args = magic_arguments.parse_argstring(_cell_magic, line)

    if args.params is not None:
        try:
            from google.cloud.bigquery.dbapi import _helpers
            import ast

            params = _helpers.to_query_parameters(ast.literal_eval("".join(args.params)))
            query_params = dict()
            for p in params:
                query_params[p.name] = p.value
            query = query.format(**query_params)
        except Exception:
            raise SyntaxError("--params is not a correctly formatted JSON string or a JSON " "serializable dictionary")
    result = _cell_magic(line, query)
    print(f"BigQuery execution took {int(time.time() - start)} seconds.")
    return result


def _load_bq_ipython_extension(ipython):
    ipython.register_magic_function(_bq_cell_magic, magic_kind="cell", magic_name="bq")


def _is_ipython():
    try:
        from IPython import get_ipython

        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


def set_gcp_credentials():
    import os
    import tempfile
    from skt.vault_utils import get_secrets

    key = get_secrets("gcp/sktaic-datahub/dataflow")["config"]
    key_file_name = tempfile.mkstemp()[1]
    with open(key_file_name, "wb") as key_file:
        key_file.write(key.encode())
        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = key_file.name


# deprecated
def import_bigquery_ipython_magic():
    load_bigquery_ipython_magic()


def load_bigquery_ipython_magic():
    if _is_ipython():
        from IPython import get_ipython

        set_gcp_credentials()
        _load_bq_ipython_extension(get_ipython())
    else:
        raise Exception("Cannot import bigquery magic. Because execution is not on ipython.")


def get_bigquery_client():
    import os
    import tempfile
    from google.cloud import bigquery
    from skt.vault_utils import get_secrets

    if "GOOGLE_APPLICATION_CREDENTIALS" in os.environ and os.path.isfile(os.environ["GOOGLE_APPLICATION_CREDENTIALS"]):
        return bigquery.Client()
    key = get_secrets("gcp/sktaic-datahub/dataflow")["config"]
    with tempfile.NamedTemporaryFile() as f:
        f.write(key.encode())
        f.seek(0)
        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = f.name
        client = bigquery.Client()
    return client


def bq_table_exists(table):
    try:
        get_bigquery_client().get_table(table)
    except NotFound:
        return False
    return True


def _get_partition_filter(dataset, table_name, partition):
    table = get_bigquery_client().get_table(f"{dataset}.{table_name}")
    if "timePartitioning" in table._properties:
        partition_column_name = table._properties["timePartitioning"]["field"]
        return f"{partition_column_name} = '{partition}'"
    elif "rangePartitioning" in table._properties:
        partition_column_name = table._properties["rangePartitioning"]["field"]
        return f"{partition_column_name} = {partition}"
    return ""


def bq_to_pandas(query, project_id="sktaic-datahub"):
    import pandas as pd

    set_gcp_credentials()
    configuration = {"query": {"useQueryCache": True}}
    return pd.read_gbq(
        query=query, project_id=project_id, dialect="standard", use_bqstorage_api=True, configuration=configuration
    )


def _bq_table_to_df(dataset, table_name, col_list, partition=None, where=None, spark_session=None):
    import base64
    from skt.vault_utils import get_secrets

    if not spark_session:
        spark_session = get_spark()
    spark_session.conf.set("spark.sql.execution.arrow.enabled", "false")
    key = get_secrets("gcp/sktaic-datahub/dataflow")["config"]
    df = (
        spark_session.read.format("bigquery")
        .option("project", "sktaic-datahub")
        .option("table", f"sktaic-datahub:{dataset}.{table_name}")
        .option("credentials", base64.b64encode(key.encode()).decode())
    )
    if partition:
        table = get_bigquery_client().get_table(f"{dataset}.{table_name}")
        if "timePartitioning" in table._properties:
            partition_column_name = table._properties["timePartitioning"]["field"]
            filter = f"{partition_column_name} = '{partition}'"
        elif "rangePartitioning" in table._properties:
            partition_column_name = table._properties["rangePartitioning"]["field"]
            filter = f"{partition_column_name} = {partition}"
        else:
            partition_column_name = None
        if partition_column_name:
            df = df.option("filter", filter)
    df = df.load().select(col_list)
    if where:
        df.where(where)
    return df


def bq_table_to_df(dataset, table_name, col_list, partition=None, where=None, spark_session=None):
    return _bq_table_to_df(dataset, table_name, col_list, partition, where, spark_session)


def bq_table_to_parquet(dataset, table_name, output_dir, col_list="*", partition=None, where=None, mode="overwrite"):
    bq_table_to_df(dataset, table_name, col_list, partition=partition, where=where).write.mode(mode).parquet(output_dir)


def bq_table_to_pandas(dataset, table_name, col_list="*", partition=None, where=None):
    col_list_str = ", ".join(col_list) if isinstance(col_list, list) else col_list
    query = f"SELECT {col_list_str} FROM {dataset}.{table_name} "
    if partition or where:
        query += "WHERE "
        conditions = []
        if partition:
            conditions.append(f" {_get_partition_filter(dataset, table_name, partition)} ")
        if where:
            conditions.append(f" {where} ")
        query += " AND ".join(conditions)
    return bq_to_pandas(query=query)


def _df_to_bq_table(df, dataset, table_name, partition=None, mode="overwrite"):
    import base64
    from skt.vault_utils import get_secrets

    key = get_secrets("gcp/sktaic-datahub/dataflow")["config"]
    table = f"{dataset}.{table_name}${partition}" if partition else f"{dataset}.{table_name}"
    df.write.format("bigquery").option("project", "sktaic-datahub").option(
        "credentials", base64.b64encode(key.encode()).decode()
    ).option("table", table).option("temporaryGcsBucket", "temp-seoul-7d").save(mode=mode)


def df_to_bq_table(df, dataset, table_name, partition=None, mode="overwrite"):
    _df_to_bq_table(df, dataset, table_name, partition, mode)


def parquet_to_bq_table(parquet_dir, dataset, table_name, partition=None, mode="overwrite"):
    try:
        spark = get_spark()
        df = spark.read.format("parquet").load(parquet_dir)
        _df_to_bq_table(df, dataset, table_name, partition, mode)
    finally:
        spark.stop()


def pandas_to_bq_table(pd_df, dataset, table_name, partition=None, mode="overwrite"):
    try:
        spark = get_spark()
        spark_df = spark.createDataFrame(pd_df)
        _df_to_bq_table(spark_df, dataset, table_name, partition, mode)
    finally:
        spark.stop()


# decorator for rdd to pandas in mapPartitions in Spark
def rdd_to_pandas(func):
    def _rdd_to_pandas(rdd_):
        import pandas as pd
        from pyspark.sql import Row

        rows = (row_.asDict() for row_ in rdd_)
        pdf = pd.DataFrame(rows)
        result_pdf = func(pdf)
        return [Row(**d) for d in result_pdf.to_dict(orient="records")]

    return _rdd_to_pandas


def bq_to_df(query, spark_session=None):
    import time

    temp_table_name = f"bq_to_df__{str(int(time.time()))}"
    temp_dataset = "temp_1d"
    jc = QueryJobConfig(
        create_disposition="CREATE_IF_NEEDED",
        write_disposition="WRITE_TRUNCATE",
        destination=f"sktaic-datahub.{temp_dataset}.{temp_table_name}",
    )
    bq_client = get_bigquery_client()
    job = bq_client.query(query, job_config=jc)
    job.result()

    return _bq_table_to_df(temp_dataset, temp_table_name, "*", spark_session=spark_session)


def load_query_result_to_table(dest_table, query, part_col_name=None, clustering_fields=None):
    bq_client = get_bigquery_client()
    qjc = None
    print(query)
    if bq_table_exists(dest_table):
        table = bq_client.get_table(dest_table)
        qjc = QueryJobConfig(
            destination=dest_table,
            write_disposition="WRITE_TRUNCATE",
            create_disposition="CREATE_IF_NEEDED",
            time_partitioning=table.time_partitioning,
            range_partitioning=table.range_partitioning,
            clustering_fields=table.clustering_fields,
        )
        job = bq_client.query(query, job_config=qjc)
        job.result()

    else:
        import time

        temp_table_name = f"load_query_result_to_table__{str(int(time.time()))}"
        bq_client.query(f"CREATE OR REPLACE TABLE temp_1d.{temp_table_name} AS {query}").result()
        if part_col_name:
            schema = bq_client.get_table(f"temp_1d.{temp_table_name}").schema
            partition_type = [f for f in schema if f.name.lower() == part_col_name.lower()][0].field_type
            if partition_type == "DATE":
                qjc = QueryJobConfig(
                    destination=dest_table,
                    write_disposition="WRITE_TRUNCATE",
                    create_disposition="CREATE_IF_NEEDED",
                    time_partitioning=TimePartitioning(field=part_col_name),
                    clustering_fields=clustering_fields,
                )
            elif partition_type == "INTEGER":
                qjc = QueryJobConfig(
                    destination=dest_table,
                    write_disposition="WRITE_TRUNCATE",
                    create_disposition="CREATE_IF_NEEDED",
                    range_partitioning=RangePartitioning(
                        PartitionRange(start=200001, end=209912, interval=1), field=part_col_name
                    ),
                    clustering_fields=clustering_fields,
                )
            else:
                print(partition_type)
                raise Exception(f"Partition column[{part_col_name}] is neither DATE or INTEGER type.")
        job = bq_client.query(f"SELECT * FROM temp_1d.{temp_table_name}", job_config=qjc)
        job.result()


def get_max_part(table_name):
    from datetime import datetime

    bq_client = get_bigquery_client()
    parts = filter(lambda x: x != "__NULL__", get_bigquery_client().list_partitions(table_name))

    if not parts:
        raise Exception("Max partition value is invalid or null.")

    max_part_value = max(parts)

    table = bq_client.get_table(table_name)
    if table.time_partitioning:
        return datetime.strptime(max_part_value, "%Y%m%d").strftime("%Y-%m-%d")
    elif table.range_partitioning:
        return int(max_part_value)
    else:
        raise Exception("Partition column is neither DATE or INTEGER type.")



__all__ = ["numeric_eda_plot", "categorical_eda_plot"]


import logging
import pandas as pd
import numpy as np
import math
import sys

from matplotlib import pyplot as plt
import seaborn as sns

sns.set_style("darkgrid", {"font.family": ["NanumGothicCoding"]})
sns.set(rc={"figure.figsize": (10, 7)})


eda_logger = logging.getLogger("EDA_TOOLS")
eda_looger = eda_logger.setLevel(logging.INFO)

stream_hander = logging.StreamHandler(sys.stdout)
stream_hander.setLevel(logging.DEBUG)

formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
stream_hander.setFormatter(formatter)
eda_logger.addHandler(stream_hander)


def numeric_eda_plot(
    df, feature_list, label_col, cols=2, n_samples=-1, plot_type="density", stat_yn=False, figsize=(7, 4),
):
    """
    Numeric feature에 대한 EDA Plot function

    Args. :
        - df           :   Pandas DataFrame 형태의 EDA대상 데이터
        - feature_list :   EDA 대상 feature list (df의 columns)
        - label_col    :   Label(or Hue) column
        - cols         :   Multi-plot 시 grid column 개수 (row 개수는 feature_list에 따라 자동으로 결정 됨)
        - n_samples    :   Label 별 sampling 할 개수 (default = -1(전수 데이터로 EDA할 경우))
        - plot_type    :   density or box (default = 'density')
        - stat_yn      :   기초 통계량 출력여부 (mean / min / max / 1q / 3q) (default : False)
        - figsize      :   (default : (7,4))

    Returns :
        matplotlib.pyplot object

    Example :
        df = load_from_hive(conn, "select * from dumbo.aips_eda_samples")
        fig = numeric_eda_plot(df, ['age','eqp_chg_cnt','real_arpu_bf_m1'], 'answer', cols = 3, n_samples = 10000,
                               plot_type='density', stat_yn=True, figsize = (7,4))
        fig
    """

    # arg type check
    assert type(feature_list) == list
    assert type(label_col) == str
    assert type(stat_yn) == bool
    assert plot_type in ["box", "density"]

    for feat in feature_list:
        assert np.issubdtype(df[feat], np.number), feat + " has %s type, should be numeric! " % (df[feat].dtype)

    ##########################################################################

    # unique lables in DF
    labels = list(df[label_col].unique())

    # label counts
    label_cnt = df.groupby(label_col).size().to_dict()

    # sampling
    if n_samples > 0:
        temp_samples_df = []
        for k, v in label_cnt.items():
            label_cnt[k] = min(n_samples, v)
            temp_samples_df.append(df[df[label_col] == k].sample(n=label_cnt[k], random_state=777))

        df = pd.concat(temp_samples_df, axis=0)

    print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
    print("LABEL COUNTS : " + str(label_cnt))
    print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
    print("")

    ##########################################################################
    # set grid for multi-plot
    if len(feature_list) == 1:
        cols = 1

    rows = math.ceil(len(feature_list) / cols)
    figwidth = figsize[0] * cols
    figheight = figsize[1] * rows

    h_margin = 0.3 + 0.2 * len(labels) if stat_yn else 0.3
    fig, ax = plt.subplots(nrows=rows, ncols=cols, figsize=(figwidth, figheight))

    color_choices = ["r", "b", "goldenrod", "black", "darkorange", "g"]
    stat_names = ["AVG", "MIN", "MAX", "1Q", "3Q"]
    pattern = "%.4f"

    plt.subplots_adjust(wspace=0.3, hspace=h_margin)
    if cols != 1 or rows != 1:
        ax = ax.ravel()  # Ravel turns a matrix into a vector... easier to iterate

    ##########################################################################
    # density plot for each ax
    for i, col in enumerate(feature_list):
        temp_ax = ax[i] if cols != 1 or rows != 1 else ax
        temp_ax.clear()
        stats = []
        x_axis_text = ""

        orig_tmp_df = df[[col] + [label_col]]

        if plot_type == "density":
            # remove outliers rows (1 percentile), 99 percentile
            tmp_df = orig_tmp_df[
                (orig_tmp_df[col] >= orig_tmp_df[col].quantile(0.01))
                & (orig_tmp_df[col] <= orig_tmp_df[col].quantile(0.99))
            ]
            for j, lb in enumerate(labels):
                sns.kdeplot(
                    tmp_df[tmp_df[label_col] == lb][col],
                    shade=True,
                    color=color_choices[j],
                    label=lb,
                    ax=temp_ax,
                    gridsize=100,
                )
                # calculate simple stats
                if stat_yn:
                    tmp_stats = [
                        np.mean(orig_tmp_df[orig_tmp_df[label_col] == lb][col]),
                        np.min(orig_tmp_df[orig_tmp_df[label_col] == lb][col]),
                        np.max(orig_tmp_df[orig_tmp_df[label_col] == lb][col]),
                        np.quantile(orig_tmp_df[orig_tmp_df[label_col] == lb][col], 0.25,),
                        np.quantile(orig_tmp_df[orig_tmp_df[label_col] == lb][col], 0.75,),
                    ]
                    tmp_stats_text = " | ".join([pattern % i for i in tmp_stats])
                    stats.append("[{}] ".format(lb) + tmp_stats_text)

        elif plot_type == "box":
            tmp_df = orig_tmp_df
            sns.boxplot(x=label_col, y=col, data=tmp_df, ax=temp_ax)
            if stat_yn:
                for j, lb in enumerate(labels):
                    tmp_stats = [
                        np.mean(orig_tmp_df[orig_tmp_df[label_col] == lb][col]),
                        np.min(orig_tmp_df[orig_tmp_df[label_col] == lb][col]),
                        np.max(orig_tmp_df[orig_tmp_df[label_col] == lb][col]),
                        np.quantile(orig_tmp_df[orig_tmp_df[label_col] == lb][col], 0.25,),
                        np.quantile(orig_tmp_df[orig_tmp_df[label_col] == lb][col], 0.75,),
                    ]
                    tmp_stats_text = " | ".join([pattern % i for i in tmp_stats])
                    stats.append("[{}] ".format(lb) + tmp_stats_text)
            temp_ax.set_ylabel("")

        if stat_yn:
            stat_title = "{}".format(" | ".join(stat_names))
            x_axis_text = stat_title + "\n" + "\n".join(stats)
        temp_ax.set_xlabel(x_axis_text)

        temp_ax.set_title(col, weight="bold")
    plt.close(fig)
    return fig


def categorical_eda_plot(df, feature_list, label_col, cols=2, n_samples=-1, figsize=(7, 4)):
    """
    Categorical feature에 대한 EDA Plot function

    Args. :
        - df           :   Pandas DataFrame 형태의 EDA대상 데이터
        - feature_list :   EDA 대상 feature list (df의 columns)
        - label_col    :   Label(or Hue) column
        - cols         :   Multi-plot 시 grid column 개수 (row 개수는 feature_list에 따라 자동으로 결정 됨)
        - n_samples    :   Label 별 sampling 할 개수 (default = -1(전수 데이터로 EDA할 경우))
        - figsize      :   (default : (7,4))

    Returns :
        matplotlib.pyplot object


    Example :
        Example :
        df = load_from_hive(conn, "select * from dumbo.aips_eda_samples")
        fig = categorical_eda_plot(df, ['family_yn', 'dangol_yn','sex_cd'], 'answer', cols = 3,
                                   n_samples = 10000, figsize = (7,4))
        fig
    """

    # arg type check
    assert type(feature_list) == list
    assert type(label_col) == str
    for feat in feature_list:
        tmp_count = df[feat].nunique()
        assert tmp_count < 100, "you have %d unique values, it is not categorical value" % (tmp_count)

    ##########################################################################

    # label counts
    label_cnt = df.groupby(label_col).size().to_dict()

    # sampling
    if n_samples > 0:
        temp_samples_df = []
        for k, v in label_cnt.items():
            label_cnt[k] = min(n_samples, v)
            temp_samples_df.append(df[df[label_col] == k].sample(n=label_cnt[k], random_state=777))

        df = pd.concat(temp_samples_df, axis=0)

    print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
    print("LABEL COUNTS : " + str(label_cnt))
    print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
    print("")

    ##########################################################################
    # set grid for multi-plot
    if len(feature_list) == 1:
        cols = 1

    rows = math.ceil(len(feature_list) / cols)
    figwidth = figsize[0] * cols
    figheight = figsize[1] * rows

    h_margin = 0.3
    fig, ax = plt.subplots(nrows=rows, ncols=cols, figsize=(figwidth, figheight))

    plt.subplots_adjust(wspace=0.3, hspace=h_margin)
    if cols != 1 or rows != 1:
        ax = ax.ravel()  # Ravel turns a matrix into a vector... easier to iterate

    ##########################################################################
    # bar plot for each ax
    for i, col in enumerate(feature_list):
        temp_ax = ax[i] if cols != 1 or rows != 1 else ax
        temp_ax.clear()
        tmp_df = df.loc[:, [col] + [label_col]]

        tmp_df = tmp_df.groupby([label_col] + [col]).size().to_frame("perc")
        tmp_df = (
            tmp_df.groupby(level=0)
            .apply(lambda x: 100 * x / float(x.sum()))
            .groupby(level=[0])
            .cumsum()
            .reset_index()
            .sort_values(["perc"], ascending=False)
        )

        sns.barplot(
            x=label_col, y="perc", hue=col, data=tmp_df, dodge=False, ax=temp_ax,
        )
        temp_ax.legend(title="")

        temp_ax.title.set_text(col)

    plt.close(fig)
    return fig



import requests
from collections import namedtuple

from skt.vault_utils import get_secrets


def get_access_token():
    secrets = get_secrets("lake/hash")
    url = secrets["auth_url"]
    client_id = secrets["client_id"]
    client_secret = secrets["client_secret"]
    data = {"grant_type": "client_credentials"}
    res = requests.post(url, auth=(client_id, client_secret), data=data)
    return res.json()["access_token"]


Task = namedtuple("Task", ["url_key", "input_key", "output_key"])
hash_task = Task("hash", "unhashedValues", "hashedValue")
unhash_task = Task("unhash", "hashedValues", "unhashedValue")


class Hash:
    access_token = get_access_token()
    url = {
        "hash": get_secrets("lake/hash")["hash_url"],
        "unhash": get_secrets("lake/hash")["unhash_url"],
    }

    @classmethod
    def renew_token(cls):
        cls.access_token = get_access_token()

    @classmethod
    def make_headers(cls):
        return {
            "Authorization": f"Bearer {cls.access_token}",
        }

    @classmethod
    def map_s(cls, values, unhash=False):
        task = hash_task
        if unhash:
            task = unhash_task
        url = cls.url[task.url_key]
        data = {"type": "s", task.input_key: values}
        r = requests.post(url, headers=cls.make_headers(), json=data)
        if r.status_code == 401:
            cls.renew_token()
            r = requests.post(url, headers=cls.make_headers(), json=data)
        if r.status_code != 200:
            raise Exception(r.content.decode("utf8"))
        return [x[task.output_key] for x in r.json()["response"]]


def hash_s(values):
    return Hash.map_s(values, unhash=False)


def unhash_s(values):
    return Hash.map_s(values, unhash=True)



def map_column(input_df, replace_column, mapping_df, _from, _to):
    """
    input_df 의 replace_column 의 값을 mapping_df 의 _from -> _to 로 map.
    mapping 된 row 과 mapping 되지 않은 row 들로 이루어진 dataframe 2 개를 리턴

    """
    mapper = mapping_df.select(_from, _to)
    output_columns = input_df.columns
    mapped = (
        input_df.join(mapper, input_df[replace_column] == mapping_df[_from], how="inner")
        .drop(replace_column)
        .withColumnRenamed(_to, replace_column)
        .select(output_columns)
    )
    remainder = input_df.join(mapper, input_df[replace_column] == mapping_df[_from], how="left_anti")
    return mapped, remainder



from collections import Counter
from skt.ye import hive_to_pandas, slack_send


# post_filter 이후, context_mapping 을 마친 post_filter_with_context_id Table 생성
def context_mapping(
    meta_table=None, comm_db=None, reco_type=None, model_name=None, dt=None, feature_ym=None,
):

    # 0. item_reco_predict_post_filter with impression 'Y' table LOAD
    query = f"""
    select *
    from {comm_db}.item_reco_predict_post_filter
    where reco_type = '{reco_type}'
    and model = '{model_name}'
    and dt = '{dt}'
    and impression_yn = 'Y'
    """
    post_filter_with_y = hive_to_pandas(query)
    print("------------------" + "predict_post_filter" + ' WITH "Y"' + " loaded")

    # 1+a. num(item_id_list)
    ITEM_ID_LIST = list(set(post_filter_with_y["prod_id"].values))
    print("------------------" + "num(item_id) : " + str(len(ITEM_ID_LIST)))

    # 1+b. item_id 한개씩 처리
    for ITEM_ID in ITEM_ID_LIST:
        print("------------------" + "item_id : " + ITEM_ID)

        # 2. context_meta Table LOAD
        query = f"""
        select reco_type, item_id, context_id, context_priority
        from {meta_table}
        where item_id = '{ITEM_ID}'
        and   reco_type = '{reco_type}'
        """
        meta = hive_to_pandas(query)
        meta = meta.drop_duplicates()
        print("------------------" + "meta_table" + ' WITH "ITEM_ID (PROD_ID)"' + " loaded")

        CONTEXT_NUM = meta.shape[0]
        CONTEXT_ID_DEFAULT = meta[meta["context_priority"] == "1"]["context_id"].values[0]
        CONTEXT_ID_LIST = tuple(meta["context_id"].values)

        # 3. default_setting
        post_filter_with_y["context_id"] = CONTEXT_ID_DEFAULT
        print("------------------" + "CONTEXT_NUM : 1 (DEFAULT)")

        if CONTEXT_NUM == 1:
            context_count_list = list(Counter(list(post_filter_with_y["context_id"].values)).items())
            msg = (
                "[CONTEXT-MAPPING-SUMMARY]"
                + " - "
                + ITEM_ID
                + "\n"
                + str(context_count_list[0][0])
                + " : "
                + str(context_count_list[0][1])
            )

        # 4. context_num >= 2 : context mapping (with context priority)
        else:
            query = f"""
            select svc_mgmt_num, dimension
            from    comm.user_profile_derivative_monthly
            where   ym = '{feature_ym}'
            and     value = 'Y'
            and     dimension in {CONTEXT_ID_LIST}
            """
            df_context = hive_to_pandas(query)

            # 4+a. for loop - overwrite mapping
            for i in range(CONTEXT_NUM - 1):
                priority_num = i + 2
                print("------------------" + "CONTEXT_NUM : " + str(priority_num))
                PRIORITY_CONTEXT_ID = meta[meta["context_priority"] == str(priority_num)]["context_id"].values[0]
                post_filter_with_y.loc[
                    (
                        post_filter_with_y["svc_mgmt_num"].isin(
                            list(df_context[df_context["dimension"] == PRIORITY_CONTEXT_ID]["svc_mgmt_num"])
                        )
                    )
                    & (post_filter_with_y["prod_id"] == ITEM_ID),
                    "context_id",
                ] = PRIORITY_CONTEXT_ID

            # 4+b. counter - context mapping statistics
            context_count_list = list(Counter(list(post_filter_with_y["context_id"].values)).items())
            msg = "[CONTEXT-MAPPING-SUMMARY]" + " - " + ITEM_ID
            for j in range(CONTEXT_NUM):
                msg_ = "\n" + str(context_count_list[j][0]) + " : " + str(context_count_list[j][1])
                msg += msg_

        del meta
        del df_context
        # 5. Slack Report
        slack_send(text=msg, channel="#rec_modeling_alert", icon_emoji=":mag:")

    # 6. return post_filter_with_y_with_context_id Table
    return post_filter_with_y
